
//Fuse CKDIV8 off und JTAGEN off, ATMEGA169P!!
//Fuses: C2, DC, FD for bootloader
//external crytal 4.194304MHz 

#include "main.h"

#define __STDIO_FDEVOPEN_COMPAT_12         // use old implementation of fdevopen
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>

const char PROGMEM digits[33]= {0x3F,0x06,0x5B,0x4F,0x66,0x6D,0x7D,0x07,0x7F,0x6F,0x40,0x20,0x08,0x01,0x58,0x10,0x50,0x58,0x3F,0x71,0x71,0x08,0x00, 0x3E, 0x79, 0x7E,0x7D,0x7B,0x77,0x6F,0x5F,0x3F,0x5C};
							//	 0	   1     2    3	   4 	5	  6    7	8	 9    -   *	    _   	  c		i	 r	  c    O	F	F	  _	 BLANK	U	  E		//LCD Tests from here

volatile char miles= 0, prev_miles, new_round=0, updated=0, HiRes2=0;	//bool variables						
volatile unsigned char oldkey, key, mode=disp_test, Display[4], delay1, delay2, delay3, blink, change, count, veloc, old_veloc, max_veloc, power, uptime=maxuptime;
volatile unsigned char hour=0, minute=0, four_second=0, port[14], oldTC, old_mode;
volatile unsigned int circum=2160, trip_100m=0, revolution, temp, trip_100m_2, tmp_rev; 
volatile unsigned long trip_mm=0, limit = 100000, temp2, limit2=100000, trip_mm2=0;		
volatile unsigned char curr_TC;

volatile unsigned long loc_tmp;

unsigned char WaitForKey (unsigned char);
int uart_putchar (char);

int main(void)
{		
	PCMSK0 |= (1<<PCINT2);					// PCI 2 on, this is PB2, reed contact wheel
	EIMSK |= (1<<PCIE0);					// Pin Change INT0 on
	ACSR |= (1<< ACD);						// Turn off analog comparator for power reduction

	OCR2A=preload;
    cli();									// disabel global interrupt
  
    TCNT2 = 0;	

    TCCR2A |= (1<<CS20) |(1<<CS22) |(1<<CS21) | (1<<WGM21);		// select precaler: 32.768 kHz / 8 = 1/64 sec between each CTC
	
	while((ASSR & 0x01) | (ASSR & 0x04));	// wait for TCN2UB and TCR2UB to be cleared
    TIFR2 = 0xFF;           				// clear interrupt-flags
    TIMSK2|= (1<<OCIE2A);     				// enable Timer2 CTC interrupt
	
	DDRA = 0xFF;
	DDRB = 0xFF; 
	DDRC = 0xFF; 
	DDRD = 0xFF;
	DDRE = 0xFB; 							//All Ports OUTPUTS except Wheel (PE2), PE0 and PE1
	DDRF = 0xEF;							//Port F4 input for external power sensing
	DDRG = 0xFF;
   
	max_veloc=0;
	power=1;
	sei();									//Enable all interrupts
	
	UCSRB |=(1 << TXEN) ;

	UBRR=6;	//38400Baud
	
	UCSRC = (3<<UCSZ0);		//8 Bit, 1 Stop bit
	fdevopen (uart_putchar, NULL, 0);
	
    while(1)                    
    {
	 
	 unsigned long tmp;
	 unsigned int time;
	 if (new_round)
	 {
		new_round=0;
		uptime = maxuptime;
		power=1;	
		trip_mm+=circum;					//add wheel circum to mmtrip
		trip_mm2+=circum;
		if (trip_mm > limit-1)
			{
			 trip_mm -= limit;				//100m are over	
			 if (++trip_100m > 9999)
				trip_100m=0;
			}
		if (trip_mm2 > limit2-1)
			{
			 trip_mm2-=limit2;
			 if (++trip_100m_2 >= 9999)
				trip_100m_2=0;
			}
		if (tmp_rev> slowest)
			veloc=0;
		else
		{
		 cli();
		   time = tmp_rev*64+curr_TC-oldTC;
		   //if ((time<280)||(time>320))
		   {
		     printf("\n%d", time);
			 printf ("  ");
			 printf (" Revs:%d", tmp_rev);
			 printf (" oldTC:%d", oldTC);
			 printf (" TC:%d", curr_TC);
			}
		   loc_tmp= circum;
		   loc_tmp= frequ*18*loc_tmp;
		   tmp=(loc_tmp+20)/40;		//2.5% of loc_tmp, rounding included
		   loc_tmp+=tmp;				//add these 2.5% to get the correct value (f=4096Hz and not 4000 which is 2.4% error)
		   loc_tmp/=time;
		   if (miles)
				veloc=(loc_tmp+4)/8;	//miles per hour, rounding included
		   else
				veloc=(loc_tmp+2)/5;	//km per hour, rounding included
		   if (veloc>max_veloc)
				max_veloc=veloc;		//store maximum speed
			sei();
		}
		oldTC=curr_TC;
     }
	 //printf("\nTick!!");
	} 
	 return (1);
}
//****************************************************************
//*************external wheel Interrupt***************************
//****************************************************************
ISR (SIG_PIN_CHANGE0)	
{
 if (!Reed_open)						//if contact closed
  {
	new_round=1;
	//OCR2A=preload;
	//while (ASSR&0x02) ;					//wait until OCR2UB reads 0, this ensures that TCNT2 reads correctly
	curr_TC = TCNT2;					//take snapshot of timer 2 status NOW!
	tmp_rev = revolution;
	//if ((!curr_TC)&&(!updated))		//in case Timer is at 0 and revolution was not updated yet -> add a revolution
	// {
	//	tmp_rev++;
	//	revolution=1;
	// }
	//else
		revolution=0;
		//printf("\n%d",tmp_rev);
  }//End of wheel interrupt....
}

//******************************************
//Timer 2 Interrupt CTC here: **************
//******************************************
ISR (TIMER2_COMP_vect)					//fires about every 15.6ms, 64Hz, takes maximum 200s to complete @8MHz CPU clock	
{	
	unsigned char i, blink_mask;

	revolution++;
	updated=1;							//signal that revolution has been added
	i= 7*(++blink & 0x01);				//can be 0 or 7 toggling every 15ms
	blink_mask = blink & 0x03;			//masks out every 16th IRQ 
	if (revolution> slowest)
	  {
		veloc=0;
		revolution= slowest+1;			//prevent speed calculation, sets speed to 0, fixed bug!!
	  }
}

//----------------------------------------------------------------------------
//############################################################################
//Routine fr die serielle Ausgabe
int uart_putchar (char c)
//############################################################################
{
	if (c == '\n')
		uart_putchar('\r');
	//Warten solange bis Zeichen gesendet wurde
	while ( !( UCSRA & (1<<UDRE)) );
	//Ausgabe des Zeichens
	UDR = c;
	return (0);
}
